#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/epoll.h>
#include <sys/sendfile.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <algorithm>
#include <map>
#include <string>
#include <utility>
#include <vector>

#include "log.h"
#include "markpool.h"
#include "mutexlock.h"
#include "parse.h"
#include "server.h"
#include "state.pb.h"
#include "stream.h"

using namespace std;

Server::Server()
{
	pthread_mutex_init(&mutex, NULL);
	pthread_mutex_init(&queued_data_mutex, NULL);

	epoll_fd = epoll_create(1024);  // Size argument is ignored.
	if (epoll_fd == -1) {
		log_perror("epoll_fd");
		exit(1);
	}
}

Server::~Server()
{
	int ret;
	do {
		ret = close(epoll_fd);
	} while (ret == -1 && errno == EINTR);

	if (ret == -1) {
		log_perror("close(epoll_fd)");
	}
}

vector<ClientStats> Server::get_client_stats() const
{
	vector<ClientStats> ret;

	MutexLock lock(&mutex);
	for (map<int, Client>::const_iterator client_it = clients.begin();
	     client_it != clients.end();
	     ++client_it) {
		ret.push_back(client_it->second.get_stats());
	}
	return ret;
}

void Server::do_work()
{
	for ( ;; ) {
		int nfds = epoll_wait(epoll_fd, events, EPOLL_MAX_EVENTS, EPOLL_TIMEOUT_MS);
		if (nfds == -1 && errno == EINTR) {
			if (should_stop) {
				return;
			}
			continue;
		}
		if (nfds == -1) {
			log_perror("epoll_wait");
			exit(1);
		}

		MutexLock lock(&mutex);  // We release the mutex between iterations.
	
		process_queued_data();

		for (int i = 0; i < nfds; ++i) {
			Client *client = reinterpret_cast<Client *>(events[i].data.u64);

			if (events[i].events & (EPOLLERR | EPOLLRDHUP | EPOLLHUP)) {
				close_client(client);
				continue;
			}

			process_client(client);
		}

		for (map<string, Stream *>::iterator stream_it = streams.begin();
		     stream_it != streams.end();
		     ++stream_it) {
			vector<Client *> to_process;
			swap(stream_it->second->to_process, to_process);
			for (size_t i = 0; i < to_process.size(); ++i) {
				process_client(to_process[i]);
			}
		}

		if (should_stop) {
			return;
		}
	}
}

CubemapStateProto Server::serialize()
{
	// We don't serialize anything queued, so empty the queues.
	process_queued_data();

	CubemapStateProto serialized;
	for (map<int, Client>::const_iterator client_it = clients.begin();
	     client_it != clients.end();
	     ++client_it) {
		serialized.add_clients()->MergeFrom(client_it->second.serialize());
	}
	for (map<string, Stream *>::const_iterator stream_it = streams.begin();
	     stream_it != streams.end();
	     ++stream_it) {
		serialized.add_streams()->MergeFrom(stream_it->second->serialize());
	}
	return serialized;
}

void Server::add_client_deferred(int sock)
{
	MutexLock lock(&queued_data_mutex);
	queued_add_clients.push_back(sock);
}

void Server::add_client(int sock)
{
	clients.insert(make_pair(sock, Client(sock)));

	// Start listening on data from this socket.
	epoll_event ev;
	ev.events = EPOLLIN | EPOLLET | EPOLLRDHUP;
	ev.data.u64 = reinterpret_cast<uint64_t>(&clients[sock]);
	if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, sock, &ev) == -1) {
		log_perror("epoll_ctl(EPOLL_CTL_ADD)");
		exit(1);
	}

	process_client(&clients[sock]);
}

void Server::add_client_from_serialized(const ClientProto &client)
{
	MutexLock lock(&mutex);
	Stream *stream;
	map<string, Stream *>::iterator stream_it = streams.find(client.stream_id());
	if (stream_it == streams.end()) {
		stream = NULL;
	} else {
		stream = stream_it->second;
	}
	clients.insert(make_pair(client.sock(), Client(client, stream)));
	Client *client_ptr = &clients[client.sock()];

	// Start listening on data from this socket.
	epoll_event ev;
	if (client.state() == Client::READING_REQUEST) {
		ev.events = EPOLLIN | EPOLLET | EPOLLRDHUP;
	} else {
		// If we don't have more data for this client, we'll be putting it into
		// the sleeping array again soon.
		ev.events = EPOLLOUT | EPOLLET | EPOLLRDHUP;
	}
	ev.data.u64 = 0;  // Keep Valgrind happy.
	ev.data.u64 = reinterpret_cast<uint64_t>(client_ptr);
	if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, client.sock(), &ev) == -1) {
		log_perror("epoll_ctl(EPOLL_CTL_ADD)");
		exit(1);
	}

	if (client_ptr->state == Client::SENDING_DATA && 
	    client_ptr->stream_pos == client_ptr->stream->bytes_received) {
		client_ptr->stream->put_client_to_sleep(client_ptr);
	} else {
		process_client(client_ptr);
	}
}

void Server::add_stream(const string &stream_id, size_t backlog_size)
{
	MutexLock lock(&mutex);
	streams.insert(make_pair(stream_id, new Stream(stream_id, backlog_size)));
}

void Server::add_stream_from_serialized(const StreamProto &stream)
{
	MutexLock lock(&mutex);
	streams.insert(make_pair(stream.stream_id(), new Stream(stream)));
}
	
void Server::set_backlog_size(const std::string &stream_id, size_t new_size)
{
	MutexLock lock(&mutex);
	assert(streams.count(stream_id) != 0);
	streams[stream_id]->set_backlog_size(new_size);
}
	
void Server::set_header(const string &stream_id, const string &header)
{
	MutexLock lock(&mutex);
	find_stream(stream_id)->header = header;

	// If there are clients we haven't sent anything to yet, we should give
	// them the header, so push back into the SENDING_HEADER state.
	for (map<int, Client>::iterator client_it = clients.begin();
	     client_it != clients.end();
	     ++client_it) {
		Client *client = &client_it->second;
		if (client->state == Client::SENDING_DATA &&
		    client->stream_pos == 0) {
			construct_header(client);
		}
	}
}
	
void Server::set_mark_pool(const std::string &stream_id, MarkPool *mark_pool)
{
	MutexLock lock(&mutex);
	assert(clients.empty());
	find_stream(stream_id)->mark_pool = mark_pool;
}

void Server::add_data_deferred(const string &stream_id, const char *data, size_t bytes)
{
	MutexLock lock(&queued_data_mutex);
	queued_data[stream_id].append(string(data, data + bytes));
}

// See the .h file for postconditions after this function.	
void Server::process_client(Client *client)
{
	switch (client->state) {
	case Client::READING_REQUEST: {
read_request_again:
		// Try to read more of the request.
		char buf[1024];
		int ret;
		do {
			ret = read(client->sock, buf, sizeof(buf));
		} while (ret == -1 && errno == EINTR);

		if (ret == -1 && errno == EAGAIN) {
			// No more data right now. Nothing to do.
			// This is postcondition #2.
			return;
		}
		if (ret == -1) {
			log_perror("read");
			close_client(client);
			return;
		}
		if (ret == 0) {
			// OK, the socket is closed.
			close_client(client);
			return;
		}

		RequestParseStatus status = wait_for_double_newline(&client->request, buf, ret);
	
		switch (status) {
		case RP_OUT_OF_SPACE:
			log(WARNING, "fd %d sent overlong request!", client->sock);
			close_client(client);
			return;
		case RP_NOT_FINISHED_YET:
			// OK, we don't have the entire header yet. Fine; we'll get it later.
			// See if there's more data for us.
			goto read_request_again;
		case RP_EXTRA_DATA:
			log(WARNING, "fd %d had junk data after request!", client->sock);
			close_client(client);
			return;
		case RP_FINISHED:
			break;
		}

		assert(status == RP_FINISHED);

		int error_code = parse_request(client);
		if (error_code == 200) {
			construct_header(client);
		} else {
			construct_error(client, error_code);
		}

		// We've changed states, so fall through.
		assert(client->state == Client::SENDING_ERROR ||
		       client->state == Client::SENDING_HEADER);
	}
	case Client::SENDING_ERROR:
	case Client::SENDING_HEADER: {
sending_header_or_error_again:
		int ret;
		do {
			ret = write(client->sock,
		                    client->header_or_error.data() + client->header_or_error_bytes_sent,
				    client->header_or_error.size() - client->header_or_error_bytes_sent);
		} while (ret == -1 && errno == EINTR);

		if (ret == -1 && errno == EAGAIN) {
			// We're out of socket space, so now we're at the “low edge” of epoll's
			// edge triggering. epoll will tell us when there is more room, so for now,
			// just return.
			// This is postcondition #4.
			return;
		}

		if (ret == -1) {
			// Error! Postcondition #1.
			log_perror("write");
			close_client(client);
			return;
		}
		
		client->header_or_error_bytes_sent += ret;
		assert(client->header_or_error_bytes_sent <= client->header_or_error.size());

		if (client->header_or_error_bytes_sent < client->header_or_error.size()) {
			// We haven't sent all yet. Fine; go another round.
			goto sending_header_or_error_again;
		}

		// We're done sending the header or error! Clear it to release some memory.
		client->header_or_error.clear();

		if (client->state == Client::SENDING_ERROR) {
			// We're done sending the error, so now close.	
			// This is postcondition #1.
			close_client(client);
			return;
		}

		// Start sending from the end. In other words, we won't send any of the backlog,
		// but we'll start sending immediately as we get data.
		// This is postcondition #3.
		client->state = Client::SENDING_DATA;
		client->stream_pos = client->stream->bytes_received;
		client->stream->put_client_to_sleep(client);
		return;
	}
	case Client::SENDING_DATA: {
sending_data_again:
		// See if there's some data we've lost. Ideally, we should drop to a block boundary,
		// but resync will be the mux's problem.
		Stream *stream = client->stream;
		size_t bytes_to_send = stream->bytes_received - client->stream_pos;
		if (bytes_to_send == 0) {
			return;
		}
		if (bytes_to_send > stream->backlog_size) {
			log(WARNING, "fd %d lost %lld bytes, maybe too slow connection",
			        client->sock,
			        (long long int)(bytes_to_send - stream->backlog_size));
			client->stream_pos = stream->bytes_received - stream->backlog_size;
			client->bytes_lost += bytes_to_send - stream->backlog_size;
			++client->num_loss_events;
			bytes_to_send = stream->backlog_size;
		}

		// See if we need to split across the circular buffer.
		bool more_data = false;
		if ((client->stream_pos % stream->backlog_size) + bytes_to_send > stream->backlog_size) {
			bytes_to_send = stream->backlog_size - (client->stream_pos % stream->backlog_size);
			more_data = true;
		}

		ssize_t ret;
		do {
			loff_t offset = client->stream_pos % stream->backlog_size;
			ret = sendfile(client->sock, stream->data_fd, &offset, bytes_to_send);
		} while (ret == -1 && errno == EINTR);

		if (ret == -1 && errno == EAGAIN) {
			// We're out of socket space, so return; epoll will wake us up
			// when there is more room.
			// This is postcondition #4.
			return;
		}
		if (ret == -1) {
			// Error, close; postcondition #1.
			log_perror("sendfile");
			close_client(client);
			return;
		}
		client->stream_pos += ret;
		client->bytes_sent += ret;

		if (client->stream_pos == stream->bytes_received) {
			// We don't have any more data for this client, so put it to sleep.
			// This is postcondition #3.
			stream->put_client_to_sleep(client);
		} else if (more_data && size_t(ret) == bytes_to_send) {
			goto sending_data_again;
		}
		break;
	}
	default:
		assert(false);
	}
}

int Server::parse_request(Client *client)
{
	vector<string> lines = split_lines(client->request);
	if (lines.empty()) {
		return 400;  // Bad request (empty).
	}

	vector<string> request_tokens = split_tokens(lines[0]);
	if (request_tokens.size() < 2) {
		return 400;  // Bad request (empty).
	}
	if (request_tokens[0] != "GET") {
		return 400;  // Should maybe be 405 instead?
	}
	if (streams.count(request_tokens[1]) == 0) {
		return 404;  // Not found.
	}

	client->stream_id = request_tokens[1];
	client->stream = find_stream(client->stream_id);
	if (client->stream->mark_pool != NULL) {
		client->fwmark = client->stream->mark_pool->get_mark();
	} else {
		client->fwmark = 0;  // No mark.
	}
	if (setsockopt(client->sock, SOL_SOCKET, SO_MARK, &client->fwmark, sizeof(client->fwmark)) == -1) {                          
		if (client->fwmark != 0) {
			log_perror("setsockopt(SO_MARK)");
		}
	}
	client->request.clear();

	return 200;  // OK!
}

void Server::construct_header(Client *client)
{
	client->header_or_error = find_stream(client->stream_id)->header;

	// Switch states.
	client->state = Client::SENDING_HEADER;

	epoll_event ev;
	ev.events = EPOLLOUT | EPOLLET | EPOLLRDHUP;
	ev.data.u64 = reinterpret_cast<uint64_t>(client);

	if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, client->sock, &ev) == -1) {
		log_perror("epoll_ctl(EPOLL_CTL_MOD)");
		exit(1);
	}
}
	
void Server::construct_error(Client *client, int error_code)
{
	char error[256];
	snprintf(error, 256, "HTTP/1.0 %d Error\r\nContent-type: text/plain\r\n\r\nSomething went wrong. Sorry.\r\n",
		error_code);
	client->header_or_error = error;

	// Switch states.
	client->state = Client::SENDING_ERROR;

	epoll_event ev;
	ev.events = EPOLLOUT | EPOLLET | EPOLLRDHUP;
	ev.data.u64 = reinterpret_cast<uint64_t>(client);

	if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, client->sock, &ev) == -1) {
		log_perror("epoll_ctl(EPOLL_CTL_MOD)");
		exit(1);
	}
}

template<class T>
void delete_from(vector<T> *v, T elem)
{
	typename vector<T>::iterator new_end = remove(v->begin(), v->end(), elem);
	v->erase(new_end, v->end());
}
	
void Server::close_client(Client *client)
{
	if (epoll_ctl(epoll_fd, EPOLL_CTL_DEL, client->sock, NULL) == -1) {
		log_perror("epoll_ctl(EPOLL_CTL_DEL)");
		exit(1);
	}

	// This client could be sleeping, so we'll need to fix that. (Argh, O(n).)
	if (client->stream != NULL) {
		delete_from(&client->stream->sleeping_clients, client);
		delete_from(&client->stream->to_process, client);
		if (client->stream->mark_pool != NULL) {
			int fwmark = client->fwmark;
			client->stream->mark_pool->release_mark(fwmark);
		}
	}

	// Bye-bye!
	int ret;
	do {
		ret = close(client->sock);
	} while (ret == -1 && errno == EINTR);

	if (ret == -1) {
		log_perror("close");
	}

	clients.erase(client->sock);
}
	
Stream *Server::find_stream(const string &stream_id)
{
	map<string, Stream *>::iterator it = streams.find(stream_id);
	assert(it != streams.end());
	return it->second;
}

void Server::process_queued_data()
{
	MutexLock lock(&queued_data_mutex);

	for (size_t i = 0; i < queued_add_clients.size(); ++i) {
		add_client(queued_add_clients[i]);
	}
	queued_add_clients.clear();	
	
	for (map<string, string>::iterator queued_it = queued_data.begin();
	     queued_it != queued_data.end();
	     ++queued_it) {
		Stream *stream = find_stream(queued_it->first);
		stream->add_data(queued_it->second.data(), queued_it->second.size());
		stream->wake_up_all_clients();
	}
	queued_data.clear();
}
